"""The update command."""

from paci.helpers.pkg_install import PkgInstall
from .base import Base


class Update(Base):
    """Updates a package!"""

    def run(self):
        install_helper = PkgInstall(self.settings, self.options, self.index, self.repo_cache, True)
        pkg_names = self.options["<package>"]

        for pkg_name in pkg_names:
            # print(self.index.get_pkg(pkg_name)) maybe add some restrictions... no new version no update?
            install_helper.install(pkg_name, self.base_pkg_dir)



import time
import sys
from halo import Halo
from termcolor import colored
from log_symbols import LogSymbols
from paci.helpers.shell import Shell

DELAY = 0.1  # time to wait for I/O to finish


def sh(cmd, cwd=None):
    """Runs a command and returns it's output

    Parameters
    ----------
    cmd : str
        The shell command which will be executed.
    cwd : str, optional
        The path to the working directory.

    Returns
    -------
    False if the execution of the command failed.
    True if the execution was successful but no output was given.
    The output of the command if it was successful and stdout wasn't empty.
    """
    executed_cmd = Shell(cmd, cwd=cwd)
    if executed_cmd.code == 0:
        if executed_cmd.stdout:
            return str(executed_cmd)
        else:
            return True
    else:
        return False


def without_keys(dictionary, keys):
    """Filters a dict and returns it. Everything in the keys list is filtered out.

    Parameters
    ----------
    dictionary : dict
        The dict which will be filtered.
    keys : list
        A list of keys as strings which will be filtered out of the dict.

    Returns
    -------
    A dict without the mentioned keys.
    """
    return {k: dictionary[k] for k in dictionary if k not in keys}


def abort_script(msg):
    """Aborts the script and outputs a message.

    Parameters
    ----------
    msg : str
        The message which will be displayed when exiting.
    """
    sys.exit("\n{} {}".format(symbol("err"), colored(msg, 'red', attrs=['bold'])))


def symbol(name):
    """Helper function to print symbols.

    Parameters
    ----------
    name : str
        You can choose a symbol by its name. These are the symbols:
           name    symbol
            []   -  ❐
            ?    -  ❓
            ok   -  ✔
            err  -  ✖
            warn -  ⚠
            info -  ℹ

    Returns
    -------
    A string containing the symbol as a Unicode char.
    """
    switcher = {
        "info": LogSymbols.INFO.value,
        "warn": LogSymbols.WARNING.value,
        "err": LogSymbols.ERROR.value,
        "ok": LogSymbols.SUCCESS.value,
        "?": colored('❓', 'red'),
        "[]": colored('❐', 'blue')
    }
    return switcher.get(name, "")


def print_heading(icon, text):
    """Helper function to print a heading.

    Parameters
    ----------
    icon : str
        Describe a name of a symbol from the symbol() function.
    text : str
        The text of the heading.
    """
    print("\n{} {}".format(symbol(icon), colored(text, 'yellow')))


def ask_with_default(question, default):
    """Helper function to get input from the use and if enter is pressed use the default.

    Parameters
    ----------
    question : str
        This is the question which will be displayed.
    default : str, optional
        You can set it to any string. It describes the default answer to the question.

    Returns
    -------
    Returns the answer to the question as a string.
    """
    return input("{} {} ({}): ".format(symbol("?"), question, default)) or default


def ask_yes_no(question, default="yes"):
    """Helper function to ask a yes/no question.

    Parameters
    ----------
    question : str
        This is the question which will be displayed.
    default : str, optional
        You can set it to "no". It describes the default answer to the question.

    Returns
    -------
    True if the answer is yes, else it returns False.
    """
    valid = {"yes": True, "y": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        choice = input("\n{} {}{}".format(symbol("?"), question, prompt)).lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no'.\n")


def run_cmd(msg, cmd, ok_msg="", err_msg="", exit_msg="", exit_all=False, cwd=None):
    """Helper function to execute a command and use spinners.

    Parameters
    ----------
    msg : str
        You should describe what the command is doing in simple words.
    cmd : list
        The command you want to execute. Each parameter can be added to the list.
    ok_msg : str, optional
        The text you want to display if the command is executed successfully.
    err_msg : str, optional
        The text you want to display if the command fails.
    exit_msg : str, optional
        The additional information you want to display if the whole script is stopped.
    exit_all : bool, optional
        Whether you want to exit the complete script if the command fails.
    cwd : str, optional
        The path to the working directory.

    Returns
    -------
    True if the command was successful else it returns False.
    """
    spinner = Halo(text=msg, spinner='dots', color='blue')
    spinner.start()
    time.sleep(DELAY)  # If this is cut out some IO operations will fail  if !(@(cmd) &>/dev/null):
    if sh(" ".join(cmd), cwd=cwd):
        spinner.succeed(colored(ok_msg if ok_msg else msg, 'green'))
        return True
    else:
        spinner.fail(colored(err_msg if err_msg else msg + " Failed!", 'red'))
        if exit_all:
            abort_script("Aborting script! " + exit_msg)
    return False


def check_version(msg, name, desired_version, current_version):
    """Checks if a version is sufficient.

    Parameters
    ----------
    msg : str
        You should describe what it is doing in simple words.
    name: str
        The name of the software which is version checked
    desired_version : tuple
        The version number in tuple formatting. (e.g. (1, 5, 4))
    current_version : str
        The version to check. (e.g. 1.5.4)
    """
    spinner = Halo(text=msg, spinner='dots', color='blue')
    spinner.start()
    time.sleep(DELAY)  # If this is cut out some IO operations will fail
    version_as_ints = (int(x) for x in current_version.split('.'))
    if tuple(version_as_ints) >= desired_version:
        spinner.succeed(colored("Version of {} is sufficient!".format(name), "green"))
    else:
        ver = ".".join([str(num) for num in desired_version])
        spinner.fail(colored("{} requires version {}!".format(name, ver), 'red'))
        abort_script("Aborting script! Please update {}!". format(name))



"""Helper to deal with the repo cache"""

import os
from tinydb import TinyDB, Query
from fuzzywuzzy import fuzz
from paci.helpers import std_helper


def find_pkg(name, repo_list, cache_path):
    """Find a package in the cached repo index files"""
    found = []
    for repo in sorted(repo_list)[::-1]:
        file = os.path.join(cache_path, "{}.json".format(repo))
        if os.path.exists(file):
            cache_db = TinyDB(file)
            res = cache_db.search(Query().name.test(fuzzy_contains, std_helper.stringify(name)))
            if res:
                for entry in res:
                    if entry not in found:
                        found.append([entry["name"], entry["ver"], entry["desc"], repo])
        else:
            print("Error! No cache found! Please use `paci refresh` first!")
            exit(1)
    if found:
        return found
    else:
        return None


def get_pkg_url(name, repo_list, cache_path):
    """Get the repo url of a package"""
    res = find_pkg(name, repo_list, cache_path)
    if res:
        return repo_list[res[0][3]]
    else:
        return None


def fuzzy_contains(val, name):
    """Fuzzy matches a string"""
    return fuzz.ratio(name, val) >= 50



"""The generate command."""

import os
from paci.helpers import file_helper
from paci.helpers.pkg_index import PkgIndex
from .base import Base


class Generate(Base):
    """Generate index files!"""

    def run(self):
        if self.options["pkg-index"]:
            print("Regenerating the index file for all installed packages...")

            file_helper.safe_delete(self.index_file)
            pkg_db = PkgIndex(self.index_file)

            pkg_count = self.generate_index(pkg_db, self.base_pkg_dir)
            print("{} packages written to: {}".format(pkg_count, self.index_file))

        if self.options["repo-index"]:
            repo_path = self.options["<path>"]
            print("Generating the index file for all packages inside a repository...")

            if os.path.isabs(repo_path):
                repo_file = os.path.join(repo_path, "index.json")

                file_helper.safe_delete(repo_file)
                repo_db = PkgIndex(repo_file)

                pkg_count = self.generate_index(repo_db, repo_path)
                print("{} packages written to: {}".format(pkg_count, repo_file))
            else:
                print("Error! Please provide an absolute path to your repository!")
                exit(1)

    @staticmethod
    def generate_index(index_db, path):
        """Generates an index file from the folders in the path."""
        pkg_count = 0

        for root, dirs, files in os.walk(path):
            files = [f for f in files if not f[0] == "."]
            dirs[:] = [d for d in dirs if not d[0] == "."]

            if "RECIPE.yml" in files:
                pkg_conf = file_helper.get_pkg_conf(os.path.join(root, "RECIPE.yml"))
                index_db.add({
                    "pkg_ver": pkg_conf["version"],
                    "pkg_desc": pkg_conf["summary"],
                    "pkg_name": pkg_conf["name"]
                })
                pkg_count += 1

        return pkg_count



import sys
from subprocess import PIPE, Popen
SUBPROCESS_HAS_TIMEOUT = True
if sys.version_info < (3, 0):
    try:
        from subprocess32 import PIPE, Popen  #NOQA
    except ImportError:
        # You haven't got subprocess32 installed. If you're running 2.X this
        # will mean you don't have access to things like timeout
        SUBPROCESS_HAS_TIMEOUT = False


class Shell(object):
    """This is lower class because it is intended to be used as a method."""

    def __init__(self, *args, **kwargs):
        self.p = None
        self.stdout = None
        self.bash(*args, **kwargs)

    def bash(self, cmd, env=None, stdout=PIPE, stderr=PIPE, timeout=None, sync=True, cwd=None):
        self.p = Popen(
            cmd, shell=True, stdout=stdout, stdin=PIPE, stderr=stderr, env=env, cwd=cwd
        )
        if sync:
            self.sync(timeout)
        return self

    def sync(self, timeout=None):
        kwargs = {'input': self.stdout}
        if timeout:
            kwargs['timeout'] = timeout
            if not SUBPROCESS_HAS_TIMEOUT:
                raise ValueError(
                    "Timeout given but subprocess doesn't support it. "
                    "Install subprocess32 and try again."
                )
        self.stdout, self.stderr = self.p.communicate(**kwargs)
        self.code = self.p.returncode
        return self

    def __repr__(self):
        return self.value()

    def __unicode__(self):
        return self.value()

    def __str__(self):
        return self.value()

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return bool(self.value())

    def value(self):
        if self.stdout:
            return self.stdout.strip().decode(encoding='UTF-8')
        return ''



"""Tests for the settings helper."""

import unittest
import os
from unittest.mock import patch, mock_open
from paci.helpers.settings import Settings

settings_data_defaults = """paci:
    temp: /tmp/paci
    base: ~/.paci
    registry:
        main: https://raw.githubusercontent.com/tradebyte/paci_packages/master
        fallback: https://raw.githubusercontent.com/tradebyte/paci_packages/master
"""

settings_data_empty = """"""

settings_data_changed = """paci:
    temp: /tmp/temp
    base: /tmp/base
    registry:
        main: https://custom.main.url
        fallback: https://custom.fallback.url
"""

settings_data_changed_partial = """paci:
    base: /tmp/base
    registry:
        fallback: https://custom.fallback.url
"""


class TestSettingsHelper(unittest.TestCase):

    @patch("paci.helpers.settings.open", mock_open(read_data=settings_data_defaults))
    def test_fetch_settings_file_defaults(self):
        """Fetching a file with default values should break nothing"""
        helper = Settings()
        settings = helper.fetch_settings()
        expected = {
            "paci": {
                "temp": "/tmp/paci",
                "base": "~/.paci",
                "registry": {
                    "main": "https://raw.githubusercontent.com/tradebyte/paci_packages/master",
                    "fallback": "https://raw.githubusercontent.com/tradebyte/paci_packages/master"
                }
            }
        }
        self.assertDictEqual(expected, settings)

    @patch("paci.helpers.settings.open", mock_open(read_data=settings_data_empty))
    def test_fetch_settings_file_empty(self):
        """Fetching an empty file should break nothing"""
        helper = Settings()
        settings = helper.fetch_settings()
        expected = {
            "paci": {
                "temp": "/tmp/paci",
                "base": os.environ.get('HOME') + '/.paci',
                "registry": {
                    "main": "https://raw.githubusercontent.com/tradebyte/paci_packages/master",
                    "fallback": "https://raw.githubusercontent.com/tradebyte/paci_packages/master"
                }
            }
        }
        self.assertDictEqual(expected, settings)

    @patch("paci.helpers.settings.open", mock_open(read_data=settings_data_changed))
    def test_fetch_settings_file_changed(self):
        """Fetching a file with all values changed should result in settings without a default value"""
        helper = Settings()
        settings = helper.fetch_settings()
        expected = {
            "paci": {
                "temp": "/tmp/temp",
                "base": "/tmp/base",
                "registry": {
                    "main": "https://custom.main.url",
                    "fallback": "https://custom.fallback.url"
                }
            }
        }
        self.assertDictEqual(expected, settings)

    @patch("paci.helpers.settings.open", mock_open(read_data=settings_data_changed_partial))
    def test_fetch_settings_file_partial(self):
        """Fetching a file with partial values changed should result in settings without some default values"""
        helper = Settings()
        settings = helper.fetch_settings()
        expected = {
            "paci": {
                "temp": "/tmp/paci",
                "base": "/tmp/base",
                "registry": {
                    "main": "https://raw.githubusercontent.com/tradebyte/paci_packages/master",
                    "fallback": "https://custom.fallback.url"
                }
            }
        }
        self.assertDictEqual(expected, settings)

    def test_fetch_settings_use_standard_path(self):
        """Check if fetch_settings uses the standard path if none is given"""
        helper = Settings()

        with patch("paci.helpers.settings.open", mock_open(read_data=settings_data_defaults)) as m:
            helper.fetch_settings()

        expected_setting_path = os.environ.get('HOME') + '/.config/paci/settings.yml'
        m.assert_called_once_with(expected_setting_path, 'r')

    def test_write_settings(self):
        """Check if we call making dirs and the file handle for writing settings"""
        towrite = {
            "paci": {
                "temp": "/custom/temp",
                "base": "/custom/base",
                "registry": {
                    "main": "https://custom.main.url",
                    "fallback": "https://custom.fallback.url"
                }
            }
        }
        helper = Settings()

        with patch("os.makedirs", return_value=False) as m_os:
            with patch("paci.helpers.settings.open", mock_open(read_data=settings_data_defaults)) as m:
                helper.write_settings(data=towrite)

        expected_setting_path = os.environ.get('HOME') + '/.config/paci/settings.yml'
        expected_setting_dir_path = os.environ.get('HOME') + '/.config/paci'
        m_os.assert_called_once_with(expected_setting_dir_path, exist_ok=True)
        m.assert_called_once_with(expected_setting_path, 'w')

    def test_settings_exist(self):
        """Check if we return false if no setting file is there, yet"""
        helper = Settings()

        with patch("os.path.isfile", return_value=False) as m:
            self.assertFalse(helper.settings_exist())



"""The refresh command."""

import os
from paci.helpers import download_helper
from .base import Base


class Refresh(Base):
    """Refreshes the repo sources list!"""

    def run(self):
        for repo in self.settings["paci"]["registry"]:
            url = os.path.join(self.settings["paci"]["registry"][repo], "index.json")
            download_helper.download(url, self.repo_cache, filename="{}.json".format(repo))

        print("\nRefreshing of the package repos complete.")



"""Helper to deal with downloads"""

import os
import stat
import subprocess


def execute_shell_script(script, working_dir):
    """Execute a shell script."""
    st = os.stat(script)
    os.chmod(script, st.st_mode | stat.S_IEXEC)
    subprocess.call(script, shell=True, cwd=working_dir)


def set_script_variables(values):
    """Sets global variables which can be used by shell scripts."""
    for key, value in values.items():
        os.environ[key] = value


def rsync(working_dir, src, dest, ignore_existing=True):
    """rsync 2 folders"""
    if ignore_existing:
        cmd = "rsync -rt --ignore-existing {}/ {}".format(src, dest)
    else:
        cmd = "rsync -rt {}/ {}".format(src, dest)
    subprocess.check_output(["bash", "-c", cmd], cwd=working_dir)


def execute(commands, working_dir):
    """Executes a string containing bash commands."""
    try:
        res = subprocess.check_output(["bash", "-c", commands], cwd=working_dir)
        for line in res.splitlines():
            print(line.decode("utf-8"))
    except subprocess.CalledProcessError as exception:
        print(exception.output)



"""The search command."""

from paci.helpers import display_helper, cache_helper
from .base import Base


class Search(Base):
    """Searches for a package!"""

    def run(self):
        pkg_name = self.options["<package>"]
        result = cache_helper.find_pkg(pkg_name, self.settings["paci"]["registry"], self.repo_cache)
        if result:
            display_helper.print_list(["Name", "Version", "Description", "Registry"], result)
        else:
            print("Package not found.")



"""Tests for our `configure` command."""

import unittest
from unittest.mock import patch, mock_open
from docopt import docopt
from io import StringIO
import paci.cli as paci
from paci.commands.configure import Configure

doc = paci.__doc__


class TestConfigure(unittest.TestCase):

    def test_prints_the_welcome_msg(self):
        configure = Configure(["test"])

        with patch('sys.stdout', new_callable=StringIO) as sysout:
            with patch("paci.commands.configure.input", return_value=False) as m:
                configure.run()

        self.assertIn('Lets configure a new settings.yml for paci!', sysout.getvalue())

    def test_asks_for_basedir(self):
        configure = Configure(["test"])

        with patch('sys.stdout', new_callable=StringIO) as sysout:
            with patch('sys.stdin', new_callable=StringIO) as sysin:
                configure.run()

        self.assertIn('basedir (defaults to', sysin.getvalue())



# -*- coding: utf-8 -*-
"""
paci

Usage:
  paci install [--no-config] [--no-cleanup] [--reuse] [--overwrite] <package>...
  paci update [--no-config] [--no-cleanup] [--reuse] [--overwrite] <package>...
  paci search <package>
  paci refresh
  paci list
  paci remove
  paci configure [--no-choice] [--silent] [--main-registry=<url>] [--fallback-registry=<url>]
  paci generate (repo-index <path> | pkg-index)
  paci --help
  paci --version

Options:
  -h, --help                         Show this screen.
  -v, --version                      Show version.
  -n, --no-config                    Omits the config.
  -c, --no-cleanup                   Don't cleanup the mess.
  -o, --overwrite                    Overwrite the config.
  -r, --reuse                        Reuse the downloaded files.
                                     (only possible with --no-cleanup)

Examples:
  paci install phpstorm

Help:
  For help using this tool, please open an issue on the Github repository:
  https://github.com/tradebyte/paci
"""

from docopt import docopt
from paci.commands import COMMANDS
from . import __version__ as version


def main():
    """Main CLI entry point."""
    import paci.commands
    options = docopt(__doc__, version=version)

    # Here we'll try to dynamically match the command the user is trying to run
    # with a pre-defined command class we've already created.
    for (key, val) in options.items():
        if hasattr(paci.commands, key) and val:
            command = COMMANDS[key](options)
            command.run()



"""Helper to deal with downloads"""

import os
import shutil
import tarfile
import ruamel.yaml
from jinja2 import Template


def create_desktop_file(values, template):
    """Creates a .desktop file for Linux and replaces the template which actual values"""
    app_dir = os.path.join(os.environ.get("HOME"), ".local/share/applications")

    # Read the file
    with open(template, "r") as file:
        desktop_file_content = Template(file.read()).render(values)

    # Write the .desktop file
    desktop_file = os.path.join(app_dir, "{}.desktop".format(values["pkg_name"]))
    with open(desktop_file, "w") as file:
        file.write(desktop_file_content)

    # Make it executable
    os.chmod(desktop_file, 0o755)


def read_yaml(file):
    """Read a YAML file and return its contents as a dict."""
    if os.path.exists(file):
        with open(file, "r") as file:
            return ruamel.yaml.load(file.read(), ruamel.yaml.RoundTripLoader)
    else:
        file(file, "w").close()
        return False


def extract_tar_gz(working_dir, file):
    """Extract a .tar.gz archive."""
    if file.endswith("tar.gz"):
        tar = tarfile.open(file, "r:gz")
        tar.extractall(working_dir)
        tar.close()


def get_pkg_conf(pkg_recipe):
    """Read the RECIPE.yml and return it."""
    if pkg_recipe:
        # Read it
        conf = read_yaml(pkg_recipe)
        if not conf:
            print("Could not read RECIPE.yml.")
            exit(1)
        else:
            return conf
    else:
        print("Abort. RECIPE.yml could not be downloaded.")
        exit(1)
    return None


def safe_delete(file):
    """Safely deletes a given file"""
    if os.path.exists(file):
        os.remove(file)


def safe_copy(src, dest):
    """Safely copies a given file to the destination"""
    os.makedirs(dest, exist_ok=True)
    shutil.copy(src, dest)



"""Helper to deal with the installation of a package"""

import tempfile
import os
import shutil
from paci.helpers import download_helper, file_helper, cmd_helper, cache_helper


class PkgInstall(object):
    """Helper for managing the installation of packages"""

    def __init__(self, settings, options, index, repo_cache, update=False):
        """Initializes the default settings."""
        self.settings = settings
        self.options = options
        self.index = index
        self.repo_cache = repo_cache
        self.update = update

    def install(self, pkg_name, base_pkg_dir):
        """Main function to walk through the installation."""
        repo_url = cache_helper.get_pkg_url(pkg_name, self.settings["paci"]["registry"], self.repo_cache)

        pkg_files = {
            "GET.json": "",
            "INSTALL.sh": "",
            "DESKTOP": "",
            "CONF.tar.gz": "",
        }

        if self.update:
            pkg_files.update({"UPDATE.sh": ""})

        if repo_url:
            pkg_url = os.path.join(repo_url, pkg_name)

            ##########################################
            # Step 1: Setup temp folder and RECIPE.yml
            ##########################################

            if self.options["--reuse"]:
                temp_dir = self.settings["paci"]["temp"]

                # Find the old temporary package directory
                sub_dirs = [name for name in os.listdir(temp_dir) if os.path.isdir(os.path.join(temp_dir, name))]
                matching_dirs = filter(lambda k: pkg_name in k, sub_dirs)
                latest_dir = max([os.path.join(temp_dir, name) for name in matching_dirs], key=os.path.getmtime)

                pkg_temp_dir = latest_dir
                pkg_recipe = os.path.join(pkg_temp_dir, "RECIPE.yml")

            else:
                # Create temporary package directory
                pkg_temp_dir = tempfile.mkdtemp(dir=self.settings["paci"]["temp"], prefix=pkg_name + "_")

                # Download RECIPE.yml
                # pylint: disable=redefined-variable-type
                # no error
                pkg_recipe = download_helper.download(os.path.join(pkg_url, "RECIPE.yml"), pkg_temp_dir, hidden=True)

            ################################
            # Step 2: Setup vars and folders
            ################################

            # Get the package configuration
            pkg_conf = file_helper.get_pkg_conf(pkg_recipe)

            # Create package directory
            pkg_dir = os.path.join(base_pkg_dir, "{}_{}".format(pkg_name, pkg_conf["version"]))
            os.makedirs(pkg_dir, exist_ok=True)

            # Create package constants (e.g. used for the templates)
            pkg_vars = {
                "pkg_src": pkg_temp_dir,
                "pkg_dir": pkg_dir,
                "pkg_ver": pkg_conf["version"],
                "pkg_desc": pkg_conf["summary"],
                "pkg_name": pkg_conf["name"]
            }

            print("Package: {} (v{})".format(pkg_vars["pkg_name"], pkg_vars["pkg_ver"]))
            print("Package working directory: {}\n".format(pkg_temp_dir))

            ###################################
            # Step 3: Download all needed files
            ###################################
            self.__download(pkg_conf, pkg_files, pkg_temp_dir, pkg_url, pkg_vars)

            ############################
            # Step 4: Start Installation
            ############################
            self.__run_install(pkg_files, pkg_vars)

            ######################
            # Step 5: Post Install
            ######################
            self.__post_install(pkg_dir, pkg_recipe, pkg_temp_dir, pkg_vars)
        else:
            print("Error! Package not found!")
            exit(1)

    def __download(self, pkg_conf, pkg_files, pkg_temp_dir, pkg_url, pkg_vars):
        """Download all needed files."""

        if self.options["--reuse"]:
            print("Reusing files...")
            # Provide path to all files needed
            for (file, _) in pkg_files.items():
                pkg_files[file] = os.path.join(pkg_temp_dir, file)
        else:
            print("Downloading files...")

            # Download all meta files
            for (file, _) in pkg_files.items():
                pkg_files[file] = download_helper.download(os.path.join(pkg_url, file), pkg_temp_dir)

            if "sources" in pkg_conf:
                pkg_files["SOURCES.tar.gz"] = download_helper.download(
                    os.path.join(pkg_url, pkg_conf["sources"]),
                    pkg_vars["pkg_src"],
                    pkg_conf["sha512sum"]
                )

            if "SOURCES.tar.gz" in pkg_files:
                file_helper.extract_tar_gz(os.path.join(pkg_vars["pkg_src"], "SOURCES"),
                                           pkg_files["SOURCES.tar.gz"])

            if pkg_files["GET.json"]:
                download_helper.download_get_files(pkg_files["GET.json"], pkg_vars["pkg_src"], pkg_vars)

    def __run_install(self, pkg_files, pkg_vars):
        """Execute the install/update script and move all files to the right place."""

        skip_install_sh = False

        if "UPDATE.sh" in pkg_files:
            if pkg_files["UPDATE.sh"]:
                print("\nUpdating package...")
                cmd_helper.set_script_variables(pkg_vars)
                cmd_helper.execute_shell_script(pkg_files["UPDATE.sh"], pkg_vars["pkg_src"])
                skip_install_sh = True

        if "INSTALL.sh" in pkg_files:
            if pkg_files["INSTALL.sh"] and not skip_install_sh:
                print("\nInstalling package...")
                cmd_helper.set_script_variables(pkg_vars)
                cmd_helper.execute_shell_script(pkg_files["INSTALL.sh"], pkg_vars["pkg_src"])

        if "DESKTOP" in pkg_files:
            if pkg_files["DESKTOP"]:
                file_helper.create_desktop_file(pkg_vars, pkg_files["DESKTOP"])

        if "CONF.tar.gz" in pkg_files and not self.options["--no-config"]:
            if pkg_files["CONF.tar.gz"]:
                conf_dir = os.path.join(pkg_vars["pkg_src"], "CONF")
                file_helper.extract_tar_gz(conf_dir, pkg_files["CONF.tar.gz"])
                cmd_helper.rsync(
                    pkg_vars["pkg_src"],
                    conf_dir, os.environ.get("HOME"),
                    False if self.options["--overwrite"] else True
                )

    def __post_install(self, pkg_dir, pkg_recipe, pkg_temp_dir, pkg_vars):
        """Handles post install tasks. Such as removing junk and adding the package to the index."""

        # Add RECIPE.yml to the pkg
        file_helper.safe_copy(pkg_recipe, pkg_dir)

        # Cleanup if successful
        if not self.options["--no-cleanup"]:
            shutil.rmtree(pkg_temp_dir)

        # Add package to the installed packages list
        self.index.add(pkg_vars)
        print("\n{} (v{}) successfully installed!\n".format(pkg_vars["pkg_name"], pkg_vars["pkg_ver"]))



"""The list command."""

import os
from paci.helpers import display_helper
from .base import Base


class List(Base):
    """Lists all installed packages!"""

    def run(self):

        if not os.path.exists(self.index_file):
            print("No packages installed yet!")
            exit(1)

        display_helper.print_list(["Name", "Version", "Description"], self.index.get_installed())



# -*- coding: utf-8 -*-
"""The base command."""

import os
from paci.helpers.settings import Settings
from paci.helpers.pkg_index import PkgIndex


class Base(object):
    """A base command."""

    def __init__(self, options, *args, **kwargs):
        """Initialize the all globally available constants."""

        self.options = options
        self.args = args
        self.kwargs = kwargs

        if not options["configure"]:
            settings_helper = Settings()

            if settings_helper.settings_exist() is False:
                settings_helper.write_settings(settings_helper.defaults)

            # Define global variables
            self.settings = settings_helper.fetch_settings()
            self.repo_cache = os.path.join(self.settings["paci"]["base"], "cache")
            self.base_pkg_dir = os.path.join(self.settings["paci"]["base"], "apps")
            self.index_file = os.path.join(self.settings["paci"]["base"], "pkgs.json")
            self.index = PkgIndex(self.index_file)

            # Ensure the directories are present
            os.makedirs(self.repo_cache, exist_ok=True)
            os.makedirs(self.base_pkg_dir, exist_ok=True)
            os.makedirs(self.settings["paci"]["temp"], exist_ok=True)

    def run(self):
        """Acts as an interface for all commands. This is where your logic goes."""
        raise NotImplementedError("You must implement the run() method yourself!")



"""The configure command lets you define a settings.yml for paci."""

import os
import json
from paci.helpers import display_helper
from paci.helpers.settings import Settings
from .base import Base


class Configure(Base):
    """Creates a settings.yml for paci interactive with the user."""

    def run(self):
        if not self.options["--no-choice"] and self.options["--silent"]:
            print("Lets configure a new settings.yml for paci!\n")

        settings_helper = Settings()

        # Get standard settings
        std_temp = settings_helper.defaults["paci"]["temp"]
        std_base = settings_helper.defaults["paci"]["base"]
        std_main = settings_helper.defaults["paci"]["registry"]["main"]
        std_fallback = settings_helper.defaults["paci"]["registry"]["fallback"]

        # Handle parameters
        def_main = std_main if not self.options["--main-registry"] else self.options["--main-registry"]
        def_fallback = std_fallback if not self.options["--fallback-registry"] else self.options["--fallback-registry"]

        if self.options["--no-choice"]:
            # Don't ask - use defaults
            if not self.options["--silent"]:
                print("Using defaults.\n")

            temp_dir = std_temp
            base_dir = std_base
            main_registry = def_main
            fallback_registry = def_fallback
        else:
            # Get user input for all values
            temp_dir = display_helper.std_input("Enter directory to store temporary files ({}):\n", std_temp)
            base_dir = display_helper.std_input("Enter file in which to save the package meta data ({}):\n", std_base)
            main_registry = display_helper.std_input("Enter main registry url ({}):\n", def_main)
            fallback_registry = display_helper.std_input("Enter fallback registry url ({}):\n", def_fallback)

        # Create settings dict
        settings = {
            "paci": {
                "temp": os.path.abspath(os.path.expanduser(temp_dir)),
                "base": os.path.abspath(os.path.expanduser(base_dir)),
                "registry": {
                    "main": main_registry,
                    "fallback": fallback_registry
                }
            }
        }

        if self.options["--no-choice"] and not self.options["--silent"]:
            print(json.dumps(settings, indent=4))

        # Save the settings
        settings_helper.write_settings(settings)



"""Helper to deal with common tasks"""


def stringify(obj):
    """Helper method which converts any given object into a string."""

    if isinstance(obj, list):
        return ''.join(obj)
    else:
        return str(obj)



"""Helper to output stuff"""

from tabulate import tabulate


def print_list(header, entries):
    """Prints out a list"""
    print(tabulate(entries, header, tablefmt="grid"))


def print_table(entries):
    """Prints out a table"""
    print(tabulate(entries, tablefmt="plain"))


def std_input(text, default):
    """Get input or return default if none is given."""
    return input(text.format(default)) or default



"""Helper class to deal with the settings.yml"""

import os
import ruamel.yaml


class Settings(object):
    """Helper for managing the paci settings.yml file"""

    def __init__(self):
        """Initializes the setting path and the default setting values."""
        self.settings = {}
        self.setting_path = os.path.join(os.environ.get("HOME"), ".config/paci/settings.yml")
        self.defaults = {
            "paci": {
                "temp": "/tmp/paci",
                "base": os.path.join(os.environ.get("HOME"), ".paci"),
                "registry": {
                    "main": "https://raw.githubusercontent.com/tradebyte/paci_packages/master",
                    "fallback": "https://raw.githubusercontent.com/tradebyte/paci_packages/master"
                }
            }
        }

    def fetch_settings(self):
        """
            Reads config from a yaml file, using default values if a setting is not defined
            in the file and returns the resulting settings as a dict.
        """
        with open(self.setting_path, "r") as file:
            try:
                # Using defaults, overwriting it with data from the given path
                settings_file = ruamel.yaml.load(file.read(), ruamel.yaml.RoundTripLoader)
            except TypeError:
                print("Provided file {} is not a valid json file! Using default settings!".format(self.setting_path))
                return self.defaults

        if settings_file is None:
            return self.defaults

        # Merging defaults and settings from file to one dict
        self.settings["paci"] = {**self.defaults["paci"], **settings_file["paci"]}
        self.settings["paci"]["registry"] = {**self.defaults["paci"]["registry"], **settings_file["paci"]["registry"]}

        # Ensure the base directory exists
        os.makedirs(os.path.dirname(self.settings["paci"]["base"]), exist_ok=True)

        return self.settings

    def write_settings(self, data):
        """Writes the given data into the settings file"""
        os.makedirs(os.path.dirname(self.setting_path), exist_ok=True)

        try:
            with open(self.setting_path, "w") as file:
                ruamel.yaml.dump(data, stream=file)
        except IOError as exception:
            print("Couldn't open or write to file (%s)." % exception)

    def settings_exist(self):
        """Returns if a setting file exists already"""
        return os.path.isfile(self.setting_path)



"""Packaging settings."""

from codecs import open
from os.path import abspath, dirname
from setuptools import find_packages, setup

from paci import __version__

this_dir = abspath(dirname(__file__))

with open("README.rst") as f:
    readme = f.read()

with open("LICENSE") as f:
    license_file = f.read()

setup(
    name="paci",
    version=__version__,
    description="Your friendly, lightweight and flexible package manager.",
    long_description=readme,
    url="https://github.com/tradebyte/paci",
    download_url="https://github.com/tradebyte/paci/archive/{}.tar.gz".format(__version__),
    author="Niklas Heer",
    author_email="niklas.heer@tradebyte.com",
    license=license_file,
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: System :: Software Distribution",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5"
    ],
    keywords=["cli", "package-manager", "install", "linux", "ubuntu"],
    packages=find_packages(exclude=["docs", "tests*"]),
    install_requires=[
        "docopt",
        "better_exceptions",
        "ruamel.yaml",
        "json-traverse",
        "clint",
        "requests",
        "jinja2",
        "tinydb",
        "tabulate",
        "fuzzywuzzy",
        "python-Levenshtein"
    ],
    entry_points={
        "console_scripts": [
            "paci=paci.cli:main",
        ],
    }
)



"""Helper to deal with downloads"""

import os
import hashlib
import requests
from clint.textui import progress
from jsontraverse.parser import JsonTraverseParser
from jinja2 import Template


def download(url, path, sha512sum=None, hidden=None, filename=None):
    """Download a file, show the progress and do integrity checks."""
    file = filename if filename else url.split("/")[-1]
    file_path = os.path.join(path, file)

    if url_exists(url):
        # Process the download and show progress
        req = requests.get(url, stream=True)
        content_length = req.headers.get("Content-Length")
        with open(file_path, "wb") as file_stream:
            # Display mode
            if content_length:
                total_length = int(content_length)
                for chunk in progress.bar(
                        req.iter_content(chunk_size=1024),
                        expected_size=(total_length / 1024) + 1,
                        label="{}: ".format(file),
                        hide=hidden
                ):
                    if chunk:
                        file_stream.write(chunk)
                        file_stream.flush()
            # Fallback Mode
            else:
                file_stream.write(requests.get(url).content)
                print("{}: [################################] 1/1 - 00:00:00".format(file))
        # Verify that the download was successful
        if sha512sum is not None:
            if not verify_file(file_path, sha512sum):
                print("{} could not be verified.".format(file))
                exit(1)

        return file_path
    else:
        return False


def download_get_files(file, pkg_src, constants):
    """Download all files from the GET.json."""
    with open(file, "r") as get_file:
        parser = JsonTraverseParser(get_file.read())
        files = parser.traverse("get")

        # work through all downloads
        for file in files:
            url = Template(file["source"]).render(constants)
            sha512sum = file["sha512sum"]
            download(url, pkg_src, sha512sum)


def url_exists(url):
    """Check if a given url exists."""
    codes = [200, 302]  # 200 OK, 302 Found
    return True if requests.head(url).status_code in codes else False


def verify_file(file, sha512sum):
    """"Returns if a file is corrupted or not."""

    # make a hash object
    hash_checker = hashlib.sha512()

    # open file for reading in binary mode
    with open(file, "rb") as file:
        # loop till the end of the file
        chunk = 0
        while chunk != b"":
            # read only 1024 bytes at a time
            chunk = file.read(1024)
            hash_checker.update(chunk)

    return hash_checker.hexdigest() == sha512sum



"""Initialize all commands which are usable in paci."""

from .install import Install
from .configure import Configure
from .update import Update
from .list import List
from .generate import Generate
from .remove import Remove
from .refresh import Refresh
from .search import Search

COMMANDS = {
    "install": Install,
    "configure": Configure,
    "update": Update,
    "list": List,
    "generate": Generate,
    "remove": Remove,
    "refresh": Refresh,
    "search": Search,
}



"""Helper to deal with the index.json"""

from tinydb import TinyDB, Query, where
from paci.helpers import std_helper


class PkgIndex(object):
    """Helper for managing the packages index.json file"""

    def __init__(self, path):
        """Initializes the setting path and the default setting values."""
        self.index_db = TinyDB(path, create_dirs=True)

    def add(self, pkg):
        """Safely adds a new record to pkg index list."""
        query = Query()

        # Create
        if not self.index_db.search(query.name == pkg["pkg_name"]):
            self.index_db.insert({"name": pkg["pkg_name"],
                                  "ver": pkg["pkg_ver"],
                                  "desc": pkg["pkg_desc"]
                                 })

        # Update
        elif self.index_db.search((query.name == pkg["pkg_name"]) & (query.ver != pkg["pkg_ver"])):
            self.index_db.update(self.__update_record(pkg), query.name == pkg["pkg_name"])

    def get_installed(self):
        """Gets a list of installed packages."""
        pkgs = []
        for pkg in self.index_db.all():
            pkgs.append([pkg["name"], pkg["ver"], pkg["desc"]])
        return pkgs

    def get_pkg(self, pkg_name):
        """Search for a package by its name."""
        return self.index_db.search(where('name') == std_helper.stringify(pkg_name))

    @staticmethod
    def __update_record(pkg):
        """Helper function to update a record."""
        def transform(element):
            """Transform the data into the right elements."""
            element["ver"] = pkg["pkg_ver"]
            element["desc"] = pkg["pkg_desc"]
        return transform
